/*
 * This file is based on work by Ross Werner.
 * Modified by Benoit Cerrina from April 2006
 * Copyright 2005, Ross Werner.
 * Distributed under the terms of the GPL version 2.
 * See LICENSE.TXT for details.
 */

package org.padict.pdb;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * @author Benoit Cerrina
 */
public class PdbFile
{
	private static int DB_NAME_LENGTH = 32;

	private String dbName;

	private boolean[] attributes;

	private int version;

	private Date createdTime;

	private Date modifiedTime;

	private Date backupTime;

	private long modifiedNumber;

	private long applicationInfoOffset;

	private long sortInfoOffset;

	private String type;

	private String creator;

	private long uniqueIdSeed;

	private long nextRecordList; // Only used internally in PalmOS

	private byte[] applicationInfoBlock;

	private byte[] sortInfoBlock;

	private List<PdbRecord> records;

	private int unknownPadding;

	private File workingFile;

	private boolean unsavedChanges;

	/* Data Accessors */

	public PdbFile()
	{
		super();
		attributes = new boolean[16];
		createdTime = new Date();
		modifiedTime = new Date();
		backupTime = new Date();
		applicationInfoBlock = new byte[0];
		sortInfoBlock = new byte[0];
		// TODO Auto-generated constructor stub
	}

	public PdbFile(boolean[] attributes, Date createdTime, Date modifiedTime, Date backupTime, long modifiedNumber, long applicationInfoOffset, long sortInfoOffset)
	{
		super();
		// TODO Auto-generated constructor stub
		this.attributes = attributes;
		this.createdTime = createdTime;
		this.modifiedTime = modifiedTime;
		this.backupTime = backupTime;
		this.modifiedNumber = modifiedNumber;
		this.applicationInfoOffset = applicationInfoOffset;
		this.sortInfoOffset = sortInfoOffset;
	}

	public String getDatabaseName()
	{
		return dbName;
	}

	// TODO: Enforce maximum length limits for all setter methods taking Strings
	public void setDatabaseName(String newDbName)
	{
		if (newDbName.length() >= DB_NAME_LENGTH)
			throw new RuntimeException("DB name length must be smaller than 32 characters");
		this.dbName = newDbName;
		this.unsavedChanges = true;
	}

	public String getCreatorName()
	{
		return creator;
	}

	public void setCreatorName(String newCreator)
	{
		this.creator = newCreator;
		this.unsavedChanges = true;
	}

	public String getTypeName()
	{
		return type;
	}

	public void setTypeName(String newType)
	{
		this.type = newType;
		this.unsavedChanges = true;
	}

	public int getVersion()
	{
		return version;
	}

	public void setVersion(int newVersion)
	{
		this.version = newVersion;
		this.unsavedChanges = true;
	}

	public long getModifiedNumber()
	{
		return modifiedNumber;
	}

	public void setModifiedNumber(long newModifiedNumber)
	{
		this.modifiedNumber = newModifiedNumber;
		this.unsavedChanges = true;
	}

	public boolean[] getAttributes()
	{
		return attributes;
	}

	public void setAttributes(boolean[] newAttributes)
	{
		attributes = newAttributes;
	}
	
	public void setAttributes(boolean[] newAttributes, int offset)
	{
		if ((newAttributes.length + offset) > attributes.length)
		{
			return;
		}

		for (int i = 0; i < newAttributes.length; i++)
		{
			attributes[i + offset] = newAttributes[i];
		}

		this.unsavedChanges = true;
	}

	public void setAttribute(boolean newAttribute, int offset)
	{
		attributes[offset] = newAttribute;
		this.unsavedChanges = true;
	}

	public Date getCreatedDateTime()
	{
		return createdTime;
	}

	public void setCreatedDateTime(Date newCreatedTime)
	{
		this.createdTime = newCreatedTime;
		this.unsavedChanges = true;
	}

	public Date getModifiedDateTime()
	{
		return modifiedTime;
	}

	public void setModifiedDateTime(Date newModifiedTime)
	{
		this.modifiedTime = newModifiedTime;
		this.unsavedChanges = true;
	}

	public Date getBackedUpDateTime()
	{
		return backupTime;
	}

	public void setBackedUpDateTime(Date newBackedUpTime)
	{
		this.backupTime = newBackedUpTime;
		this.unsavedChanges = true;
	}

	public PdbRecord getRecord(int index)
	{
		return (PdbRecord) records.get(index);
	}

	public List<PdbRecord> getRecords()
	{
		if (records == null)
			records = new ArrayList<PdbRecord>();
		return records;
	}

	public PdbRecord createNewRecord()
	{
		if (applicationInfoOffset > 0)
			applicationInfoOffset += 8;
		if (sortInfoOffset > 0)
			sortInfoOffset += 8;

		PdbRecord record = (PdbRecord) records.get(records.size() - 1);

		// Find a unique ID
		// FIXME: Clean up code
		int uniqueID = record.getId() + 1;
		boolean idIsUnique;
		do
		{
			idIsUnique = true;
			Iterator it = records.iterator();
			while (it.hasNext())
			{
				record = (PdbRecord) it.next();
				if (record.getId() == uniqueID)
				{
					idIsUnique = false;
					uniqueID++;
					break;
				}
			}
		} while (!idIsUnique);

		PdbRecord newRecord = new PdbRecord(new byte[0], (byte) 0x00, uniqueID);
		records.add(newRecord);

		return newRecord;
	}

	public void deleteRecord(int recordNumber)
	{
		if (applicationInfoOffset > 0)
			applicationInfoOffset -= 8;
		if (sortInfoOffset > 0)
			sortInfoOffset -= 8;

		records.remove(recordNumber);
	}

	public void setWorkingFile(File newWorkingFile)
	{
		this.workingFile = newWorkingFile;
	}

	public boolean unsavedChanges()
	{
		return unsavedChanges;
	}

	/* Utility Methods */

	public boolean save() throws IOException
	{
		FileOutputStream output = new FileOutputStream(workingFile);
		writeString(output, dbName, 32);
		writeAttributes(output, attributes);
		writeWord(output, version);
		writeDateTime(output, createdTime);
		writeDateTime(output, modifiedTime);
		writeDateTime(output, backupTime);
		writeDWord(output, modifiedNumber);
		writeDWord(output, applicationInfoOffset);
		writeDWord(output, sortInfoOffset);
		writeString(output, type, 4);
		writeString(output, creator, 4);
		writeDWord(output, uniqueIdSeed);
		writeDWord(output, nextRecordList);
		writeWord(output, records.size());

		int offset = 0x4e;
		offset += (records.size() * 8);
		offset += applicationInfoBlock.length;
		offset += sortInfoBlock.length;
		offset += unknownPadding;

		// Write Record Headers (not records themselves)
		for (int i = 0; i < records.size(); i++)
		{
			PdbRecord record = (PdbRecord) records.get(i);
			writeDWord(output, offset);
			writePartialWord(output, record.getAttributes(), 1);
			writePartialWord(output, record.getId(), 3);

			offset += record.getSize();
		}

		byte[] padding = new byte[unknownPadding];
		for (int i = 0; i < padding.length; i++)
		{
			padding[i] = 0x00;
		}
		output.write(padding);

		output.write(applicationInfoBlock);

		for (int i = 0; i < records.size(); i++)
		{
			PdbRecord record = (PdbRecord) records.get(i);
			output.write(record.getData());
		}

		output.close();
		return true;
	}

	/* Private Methods */

	private void writeString(FileOutputStream output, String string, int numBytes) throws IOException
	{
		byte[] bytes = new byte[numBytes];
		byte[] partial = string.getBytes();

		for (int i = 0; i < numBytes; i++)
		{
			if (i < partial.length)
			{
				bytes[i] = partial[i];
			} else
			{
				bytes[i] = 0x00;
			}
		}

		output.write(bytes);
	}

	private void writeAttributes(FileOutputStream output, boolean[] attributes) throws IOException
	{
		byte[] bytes = new byte[attributes.length / 8];

		for (int i = 0; i < bytes.length; i++)
		{
			bytes[i] = 0;
		}

		for (int i = 0; i < attributes.length; i++)
		{
			if (attributes[i])
			{
				bytes[(bytes.length - 1) - (i / 8)] |= (byte) (0x1 << (i % 8));
			}
		}

		output.write(bytes);
	}

	private void writeWord(FileOutputStream output, int word) throws IOException
	{
		byte[] buffer = new byte[2];
		buffer[0] = (byte) ((word >> 8) & 0xff);
		buffer[1] = (byte) (word & 0xff);

		output.write(buffer);
	}

	private void writePartialWord(FileOutputStream output, int word, int numBytes) throws IOException
	{
		byte[] buffer = new byte[numBytes];

		for (int i = 0; i < numBytes; i++)
		{
			buffer[i] = (byte) (word >> ((numBytes - i - 1) * 8) & 0xff);
		}

		output.write(buffer);
	}

	private void writeDateTime(FileOutputStream output, Date date) throws IOException
	{
		long dateTime = (date.getTime() / 1000) + 2082844800L;
		writeDWord(output, dateTime);
	}

	private void writeDWord(FileOutputStream output, long dword) throws IOException
	{
		byte[] buffer = new byte[4];
		buffer[0] = (byte) ((dword >> 24) & 0xff);
		buffer[1] = (byte) ((dword >> 16) & 0xff);
		buffer[2] = (byte) ((dword >> 8) & 0xff);
		buffer[3] = (byte) (dword & 0xff);
		output.write(buffer);
	}

	public String toString()
	{
		String toReturn = "";

		toReturn += "dbName:               " + dbName + "\n";
		toReturn += "attributes:           ";

		for (int i = 0; i < 16; i++)
		{
			toReturn += attributes[i] ? "x" : ".";
		}
		toReturn += "\n";

		toReturn += "version:              " + version + "\n";
		toReturn += "createdTime:          " + createdTime + "\n";
		toReturn += "modifiedTime:         " + modifiedTime + "\n";
		toReturn += "backupTime:           " + backupTime + "\n";
		toReturn += "modifiedNumber:       " + modifiedNumber + "\n";
		toReturn += "appInfoSize:          " + applicationInfoOffset + "\n";
		toReturn += "sortInfoSize:         " + sortInfoOffset + "\n";
		toReturn += "type:                 " + type + "\n";
		toReturn += "creator:              " + creator + "\n";
		toReturn += "uniqueID:             0x" + Integer.toHexString((int) uniqueIdSeed) + "\n";
		toReturn += "nextRecord:           " + nextRecordList + "\n";
		toReturn += "numberOfRecords:      " + records.size() + "\n";

		for (int i = 0; i < records.size(); i++)
		{
			toReturn += "\n";
			toReturn += records.get(i).toString();
		}

		return toReturn;
	}
}